/*
** shortkeys.c
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2.1 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <nokix/all.h>
#include "global.h"

#define TIME_HOLD             40
#define TIME_WAIT_KEY         375

#define OMSG_FIREUP           (0x1e00)
#define OMSG_TIMER_HOLD       (0x1e01)
#define OMSG_TIMER_WAIT_KEY   (0x1e02)

static unsigned char xstatus, group_key;

#define STATE_IDLE            0
#define STATE_HOLD            1
#define STATE_WAIT_RELEASE    2
#define STATE_WAIT_KEY        3
#define STATUS_STATE_MASK     0x0f

#define FLAG_CLOSE_MSG        0x80
#define STATUS_FLAGS_MASK     0xf0

#define xstate (xstatus&STATUS_STATE_MASK)
#define xflags (xstatus&STATUS_FLAGS_MASK)
#define xstate_set(s) xstatus=xflags|s
#define xflags_set(f) xstatus|=f
#define xflags_clr(f) xstatus&=~f

/**************************************************************************************/

static const unsigned timer_hold=OMSG_TIMER_HOLD;
static const unsigned timer_wait_key=OMSG_TIMER_WAIT_KEY;

extern const ringtone_t tone;
extern unsigned char menu_shortcut_state;

extern rexx_uint fireup_handler_id;

/**************************************************************************************/

/* converts system key code to shortkey code */
static unsigned get_keycode(void)
{
   unsigned code=msg_argv[0];
   
   if(code==KEY_EXTENDED) code=0x80|msg_argv[1];
   return code;
}

/* converts key id to shortkeys table index */
static int get_index(unsigned group_key, unsigned key, unsigned flags, unsigned mask)
{
   unsigned i;
   struct shortkey *s;
   
   for(i=0;shortkeys[i].key;i++)
   {
      s=&shortkeys[i];
      
      if(s->group_key==group_key)
      {
         if((key==0)||(s->key==key))
         {
            if(keylock_status==KEYLOCK_STATUS_UNLOCKED)
            {
               if(s->flags&MF_LOCKED) continue;
            }
            else if(keylock_status==KEYLOCK_STATUS_LOCKED)
            {
               if(!(s->flags&(MF_ALWAYS|MF_LOCKED))) continue;
            }
            else
            {
               if(!(s->flags&MF_ALWAYS)) continue;
            }
            
            if( netmonitor_status != 0 && ( s->flags & MF_NONMON ) ) continue;
            if( dial_number_buf[0] != '\0' && ( s->flags & MF_NODIAL ) ) continue;

            if((s->flags&mask)==(flags&mask)) return i;
         }
      }
   }

   return -1;
}

/* clears the dial number buffer and fires up a shortkey */
static void fireup(int idx)
{
   if(dial_number_buf[0])
   {
      // clear number buffer
      dial_number_buf[0]=0;

      // send clear key to refresh screen
      send_message(MSG_KEYPRESS|MSG_1P,KEY_CLEAR);
      send_message(MSG_KEYRELEASE|MSG_1P,KEY_CLEAR);
   }
   
   if((keylock_status==KEYLOCK_STATUS_LOCKED)&&
      ((shortkeys[idx].flags)&MF_UNLOCK))
   {
      send_message(0x6CC);
      xflags_set(FLAG_CLOSE_MSG);
   }

   // fireup
   send_message(OMSG_FIREUP|MSG_1P,idx);

   // reset state
   xstate_set(STATE_IDLE);
}

/**************************************************************************************/

/* shortkeys messages dispatcher */
unsigned shortkeys_handler(unsigned msg)
{
   if( current_handler != 1 || get_phone_state() != PHST_POWER_ON_NORMAL ) return msg;

   switch(msg)
   {
      case MSG_KEYPRESS:
         {
            int i;
            
            if(xstate==STATE_WAIT_KEY)
            {
               own_timer_abort(&timer_wait_key);
               xstate_set(STATE_IDLE);

               i=get_index(group_key,get_keycode(),0,0);
               if(i!=-1)
               {
                  fireup(i);
                  msg_argv[0]=KEY_NONE;
               }
            }
            else
            {
               group_key=get_keycode();

               // check direct shortkey (down)
               i=get_index(0,group_key,MF_DOWN,MF_DOWN|MF_UP);
               if(i!=-1)
               {
                  fireup(i);
                  msg_argv[0]=KEY_NONE;
                  break;
               }

               // check direct shortkey (up)
               i=get_index(0,group_key,MF_UP,MF_DOWN|MF_UP);
               if(i!=-1)
               {
                  // this will be catched in MSG_KEYRELEASE but we have to
                  // disable original function of the key
                  msg_argv[0]=KEY_NONE;
               }

               // check hold shortkey
               i=get_index(0,group_key,0,MF_DOWN|MF_UP);
               if(i!=-1)
               {
                  xstate_set(STATE_HOLD);
                  own_timer_start(&timer_hold,TT_MESSAGE,TIME_HOLD);
                  break;
               }

               // check group key
               i=get_index(group_key,0/*= any key*/,0,MF_UP);
               if(i!=-1)
               {
                  if((shortkeys[i].flags)&MF_DOWN)
                  {
                     // direct group key (down)
                     xstate_set(STATE_WAIT_KEY);
                     own_timer_start(&timer_wait_key,TT_MESSAGE,TIME_WAIT_KEY);
                  }
                  else
                  {
                     // hold group key
                     xstate_set(STATE_HOLD);
                     own_timer_start(&timer_hold,TT_MESSAGE,TIME_HOLD);
                  }
               }
            }
         }
         break;

      case MSG_KEYRELEASE:
         if(xstate==STATE_HOLD)
         {
            own_timer_abort(&timer_hold);
            xstate_set(STATE_IDLE);
         }
         else if(xstate==STATE_WAIT_RELEASE)
         {
            own_timer_start(&timer_wait_key,TT_MESSAGE,TIME_WAIT_KEY);
            xstate_set(STATE_WAIT_KEY);
         }

         if(xstate==STATE_IDLE)
         {
            int i;

            // check direct shortkey (up)
            i=get_index(0,get_keycode(),MF_UP,MF_DOWN|MF_UP);
            if(i!=-1)
            {
               fireup(i);
               msg=MSG_NOP;
               break;
            }

            // check group key
            i=get_index(group_key,0/*= any key*/,MF_UP,MF_DOWN|MF_UP);
            if(i!=-1)
            {
               // direct group key (up)
               xstate_set(STATE_WAIT_KEY);
               own_timer_start(&timer_wait_key,TT_MESSAGE,TIME_WAIT_KEY);
            }
         }
         break;

      case OMSG_TIMER_HOLD:
         {
            int i;

            // check hold shortkey
            i=get_index(0,group_key,0,MF_DOWN|MF_UP);
            if(i!=-1)
            {
               fireup(i);
            }
            else
            {
               // group key
               stop_tone(NULL,2); // stop the keypad tone
               own_play_tone(&tone,PTM_WARNING_TONE);
               xstate_set(STATE_WAIT_RELEASE);
            }

            msg=MSG_NOP;
         }
         break;

      case OMSG_TIMER_WAIT_KEY:
         xstate_set(STATE_IDLE);
         msg=MSG_NOP;
         break;

      case OMSG_FIREUP:
         {
            unsigned key=msg_argv[0];
            struct shortkey *s=&shortkeys[key];

            send_message(MSG_CLOSE_DISPLAY_MESSAGE);

            if(s->flags&MF_THROUGH_HANDLER) send_message(MSG_START_HANDLER|MSG_2P,fireup_handler_id,key);
            else send_message(s->msg|MSG_1P,s->arg);

            if(s->flags&MIF_FINISH_SHORTCUT) menu_shortcut_state=0;
            else menu_shortcut_state=2; /* enable shortcuts */

            if(s->flags&MIF_SHOW_DONE) send_message(MSG_DISPLAY_MESSAGE|MSG_2P,DMI_DONE,str_done);

            msg=MSG_NOP;
         }
         break;
         
      case 0x6b3: // shows "Keypad active" on newer models (handled in ui_preprocess)
         if(xflags&FLAG_CLOSE_MSG)
         {
            xflags_clr(FLAG_CLOSE_MSG);
            msg=MSG_NOP;
         }
         break;
   }
   return msg;
}

/**************************************************************************************/

/*
EXTENDED 5510 CODES:

S			(0x01)1
D			(0x02)2
G			(0x03)3
W			(0x04)4
F			(0x05)5
E			(0x06)6
A			(0x07)7
!        8
!        9
T			(0x0a)10
!        11
R			(0x0c)12
!        13
Q			(0x0e)14
!        15
!        16
!        17
!        18
Y			(0x13)19
!        20
!        21
P        (0x16)22
O			(0x17)23
N			(0x18)24
J			(0x19)25
U			(0x1a)26
I			(0x1b)27
H			(0x1c)28
:	      (0x1d)29
L			(0x1e)30
,    		(0x1f)31
?        (0x20)32
K			(0x21)33
.		   (0x22)34
M			(0x23)35
+    		(0x24)36
*        (0x25)37
B			(0x26)38
X			(0x27)39
V			(0x28)40
C			(0x29)41
Z			(0x2a)42
SPACE    (0x2b)43

*/

